package com.fortanix.sdkms.demo;

import org.apache.commons.codec.binary.Base64;

import javax.crypto.Cipher;
import javax.crypto.KeyGenerator;
import javax.crypto.SecretKey;
import javax.crypto.spec.IvParameterSpec;
import java.io.Console;
import java.nio.charset.Charset;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.StandardOpenOption;
import java.security.*;
import java.util.Arrays;
import java.util.List;

/**
 * Created by anand on 4/3/17.
 * Edited: 20/05/2022
 */
public class Sample {

    private static Provider initialize() {
        try {
            final Path tmpFile = Files.createTempFile("pkcs11", ".cfg");
            tmpFile.toFile().deleteOnExit();
            final List<String> lines = Arrays.asList("name = fortanix", "library = /opt/fortanix/pkcs11/fortanix_pkcs11.so");
            Files.write(tmpFile, lines, Charset.defaultCharset(), StandardOpenOption.WRITE);

            Provider provider = Security.getProvider("SunPKCS11");
            provider = provider.configure(tmpFile.toString());
            Security.addProvider(provider);

            System.out.println("Successfully initialized JCE to use Fortanix SDKMS");
            return provider;
        } catch (Exception e) {
            System.out.println("Could not initialize: " + e);
            return null;
        }
    }

    private static boolean login(Provider provider, String apiKey) {;
        try {
            KeyStore ks = KeyStore.getInstance("PKCS11", provider);
            ks.load(null, apiKey.toCharArray());
            System.out.println("Successfully logged in to Fortanix SDKMS");
            return true;
        } catch (Exception e) {
            System.out.println("Incorrect API Key: " + e);
            return false;
        }
    }

    private static boolean logout(Provider provider) {
        try {
            ((AuthProvider) provider).logout();
            return true;
        } catch (Exception e) {
            System.out.println("Can't logout: " + e);
            return false;
        }
    }

    private static SecretKey generateKey(Provider provider, String type, int length) {
        try {
            KeyGenerator keyGen = KeyGenerator.getInstance(type, provider);
            keyGen.init(length);
            SecretKey key = keyGen.generateKey();
            System.out.println("Successfully generated secret key of type " + type + " and length " + length);
            return key;
        } catch (Exception e) {
            System.out.println("Can't generate secret key: " + e);
            return null;
        }
    }

    private static String encrypt(Provider provider, SecretKey key, String plain, String mode, int length) {
        byte[] iv = new byte[length / 8];
        SecureRandom prng = new SecureRandom();
        prng.nextBytes(iv);

        try {
            Cipher cipher = Cipher.getInstance(mode, provider);
            cipher.init(Cipher.ENCRYPT_MODE, key, new IvParameterSpec(iv));
            byte[] byteCipherText = cipher.doFinal(plain.getBytes());
            String cryptStruct = new Base64().encodeAsString(iv) + ":" + new Base64().encodeAsString(byteCipherText);
            return  cryptStruct;
        } catch (Exception e) {
            System.out.println("Encryption failed: " + e);
            return null;
        }
    }

    private static String decrypt(Provider provider, SecretKey key, String cipherStruct, String mode) {
        try {
            String[] ivAndCipherText = cipherStruct.split(":");
            byte[] iv = new Base64().decode(ivAndCipherText[0]);
            byte[] cipherText = new Base64().decode(ivAndCipherText[1]);
            Cipher cipher = Cipher.getInstance(mode, provider);
            cipher.init(Cipher.DECRYPT_MODE, key, new IvParameterSpec(iv));
            byte[] byteCipherText = cipher.doFinal(cipherText);

            return new String(byteCipherText);
        } catch (Exception e) {
            System.out.println("Decryption failed: " + e);
            return null;
        }
    }

    public static void main(String[] args) throws InterruptedException {
        final String strDataToEncrypt = "The quick brown fox jumps over the lazy dog";
        final int AES_KEYLENGTH = 128;  // change this as desired for the security level you want
        final String keyType = "AES";
        final String mode = "AES/CBC/PKCS5Padding";
        final String pin;

        if (args.length == 1) {
            pin = args[1];
        } else {
            final Console console = System.console();
            pin = console.readLine("Enter API Key: ");
        }

        final Provider provider = initialize();
        if (provider == null) System.exit(1);

        if (!login(provider, pin)) System.exit(1);

        final SecretKey secretKey = generateKey(provider, keyType, AES_KEYLENGTH);
        if (secretKey == null) System.exit(1);

        System.out.println("Encrypting string: " + strDataToEncrypt);
        final String cipherStruct = encrypt(provider, secretKey, strDataToEncrypt, mode, AES_KEYLENGTH);
        if (cipherStruct == null) System.exit(1);
        System.out.println("Encrypted string: " + cipherStruct);

        final String decryptText = decrypt(provider, secretKey, cipherStruct, mode);
        if (decryptText == null) System.exit(1);
        System.out.println("Decrypted string: " + decryptText);

        if (!logout(provider)) System.exit(1);
    }
}
